package xyz.lychee.lagfixer.libs.adventure.pointer;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;

final class PointersSupplierImpl<T> implements PointersSupplier<T> {
   private final PointersSupplier<? super T> parent;
   private final Map<Pointer<?>, Function<T, ?>> resolvers;

   PointersSupplierImpl(@NotNull final @NotNull BuilderImpl<T> builder) {
      this.parent = builder.parent;
      this.resolvers = new HashMap(builder.resolvers);
   }

   public @NotNull Pointers view(final @NotNull T instance) {
      return new ForwardingPointers(instance, this);
   }

   public <P> boolean supports(final @NotNull Pointer<P> pointer) {
      if (this.resolvers.containsKey(Objects.requireNonNull(pointer, "pointer"))) {
         return true;
      } else {
         return this.parent == null ? false : this.parent.supports(pointer);
      }
   }

   public <P> @Nullable Function<? super T, P> resolver(final @NotNull Pointer<P> pointer) {
      Function<? super T, ?> resolver = (Function)this.resolvers.get(Objects.requireNonNull(pointer, "pointer"));
      if (resolver != null) {
         return resolver;
      } else {
         return this.parent == null ? null : this.parent.resolver(pointer);
      }
   }

   static final class BuilderImpl<T> implements PointersSupplier.Builder<T> {
      private PointersSupplier<? super T> parent = null;
      private final Map<Pointer<?>, Function<T, ?>> resolvers = new HashMap();

      @NotNull
      public PointersSupplier.@NotNull Builder<T> parent(final @Nullable PointersSupplier<? super T> parent) {
         this.parent = parent;
         return this;
      }

      @NotNull
      public <P> PointersSupplier.@NotNull Builder<T> resolving(final @NotNull Pointer<P> pointer, final @NotNull Function<T, P> resolver) {
         this.resolvers.put(pointer, resolver);
         return this;
      }

      public @NotNull PointersSupplier<T> build() {
         return new PointersSupplierImpl(this);
      }
   }

   static final class ForwardingPointers<U> implements Pointers {
      private final U instance;
      private final PointersSupplierImpl<U> supplier;

      ForwardingPointers(final @NotNull U instance, final @NotNull PointersSupplierImpl<U> supplier) {
         this.instance = instance;
         this.supplier = supplier;
      }

      public <T> @NotNull Optional<T> get(final @NotNull Pointer<T> pointer) {
         Function<? super U, ?> resolver = (Function)this.supplier.resolvers.get(Objects.requireNonNull(pointer, "pointer"));
         if (resolver == null) {
            resolver = this.supplier.parent.resolver(pointer);
         }

         return resolver == null ? Optional.empty() : Optional.ofNullable(resolver.apply(this.instance));
      }

      public <T> boolean supports(final @NotNull Pointer<T> pointer) {
         return this.supplier.supports(pointer);
      }

      public Pointers.@NotNull Builder toBuilder() {
         Pointers.Builder builder = this.supplier.parent == null ? Pointers.builder() : (Pointers.Builder)this.supplier.parent.view(this.instance).toBuilder();
         Iterator var2 = this.supplier.resolvers.entrySet().iterator();

         while(var2.hasNext()) {
            Map.Entry<Pointer<?>, Function<U, ?>> entry = (Map.Entry)var2.next();
            builder.withDynamic((Pointer)entry.getKey(), () -> {
               return ((Function)entry.getValue()).apply(this.instance);
            });
         }

         return builder;
      }
   }
}
